use crate::aes::*;
use rand::{Rng, thread_rng};

impl Aes {
    pub fn reverse_key_schedule(original_round_key: &[u8; 16], aes_round: usize) -> [u8; 16] {
        let mut rcon = [0u8; 4];
        let mut round_key: [u8; 16] = [0u8; 16];
        round_key.copy_from_slice(original_round_key);

        for i in (0..aes_round).rev() {
            // println!("{}", i);
            rcon[0] = RC[i + 1];
            // println!("{:02x}", rcon[0]);
            let a2: [u8; 4] = round_key[0..4].try_into().unwrap();
            let b2: [u8; 4] = round_key[4..8].try_into().unwrap();
            let c2: [u8; 4] = round_key[8..12].try_into().unwrap();
            let d2: [u8; 4] = round_key[12..16].try_into().unwrap();

            let d1 = xor(&d2, &c2);
            let c1 = xor(&c2, &b2);
            let b1 = xor(&b2, &a2);
            let a1 = xor(&xor(&a2, &shift_word(&substitute_word(&d1))), &rcon);

            round_key[0..4].copy_from_slice(&a1);
            round_key[4..8].copy_from_slice(&b1);
            round_key[8..12].copy_from_slice(&c1);
            round_key[12..16].copy_from_slice(&d1);
        }
        round_key
    }

    pub fn aes_reduced(&key: &[u8; 16], &block: &[u8; 16]) -> [u8; 16] {
        let nturn = 4;
        let aescipher = Self::new(&key, &nturn);
        aescipher.encrypt_block(&block)
    }
    pub fn aes_reduced_gen_texts(&key: &[u8; 16], &const_byte: &u8) -> [[u8; 16]; 256] {
        let mut cleartext: [u8; 16];
        let mut ciphertexts: [[u8; 16]; 256] = [[0; 16]; 256];
        for i in 0..=255 {
            cleartext = [
                i, const_byte, const_byte, const_byte, const_byte, const_byte, const_byte,
                const_byte, const_byte, const_byte, const_byte, const_byte, const_byte, const_byte,
                const_byte, const_byte,
            ];
            ciphertexts[i as usize] = Aes::aes_reduced(&key, &cleartext);
        }
        ciphertexts
    }
    pub fn decrypt_block_reduced_1_step(&self, block: &[u8; 16]) -> [u8; 16] {
        let mut result = [0u8; 16];

        let mut state = [[0u8; 4]; 4];
        for i in 0..16 {
            state[i % 4][i / 4] = block[i];
        }
        add_round_key(&mut state, &clone_into_array(&self.expanded_key[16..20]));
        inverse_shift_rows(&mut state);
        inverse_substitute_state(&mut state);

        for i in 0..4 {
            for j in 0..4 {
                result[4 * j + i] = state[i][j]
            }
        }

        result
    }
    pub fn guessroundkey(&texts: &[[u8; 16]; 256]) -> [Vec<u8>; 16] {
        let mut key: [Vec<u8>; 16] = Default::default();
        let mut s: u8;
        let mut matrix: [[u8; 4]; 4] = [[0; 4]; 4];
        let mut array: [u8; 16] = [0; 16];
        for j in 0..16 {
            for i in 0..=255 {
                s = 0x00;
                for ciphertext in texts {
                    for i in 0..4 {
                        for j in 0..4 {
                            matrix[j][i] = ciphertext[j * 4 + i];
                        }
                    }
                    for i in 0..4 {
                        for j in 0..4 {
                            array[j * 4 + i] = matrix[j][i];
                        }
                    }
                    s ^= substitute(array[j] ^ i, false);
                }
                if s == 0x00 {
                    println!("found: {:02x} for byte {}", i, j);
                    key[j].push(i);
                }
            }
            println!();
        }
        key
    }

    pub fn findroundkey(&key: &[u8; 16]) -> [u8; 16] {
        let mut found_key: [u8; 16] = [0x00; 16];
        let mut const_byte: u8 = 0x00;
        let mut ciphertexts: [[u8; 16]; 256] = Self::aes_reduced_gen_texts(&key, &const_byte);
        let mut key_guesses: [Vec<u8>; 16] = Self::guessroundkey(&ciphertexts);
        let mut more_key_guesses: [Vec<u8>; 16];
        let mut result: [Vec<u8>; 16] = Default::default();
        while key_guesses.iter().any(|vec| vec.len() > 1) {
            println!("Collisions present, additional key search with a different constant.");
            const_byte += 1_u8;
            ciphertexts = Self::aes_reduced_gen_texts(&key, &const_byte);
            more_key_guesses = Self::guessroundkey(&ciphertexts);
            for ((vec1, vec2), result_vec) in more_key_guesses
                .iter()
                .zip(key_guesses.iter())
                .zip(result.iter_mut())
            {
                let common_elements: Vec<u8> = vec1
                    .iter()
                    .filter(|&byte| vec2.contains(byte))
                    .copied()
                    .collect();

                *result_vec = common_elements;
            }
            for (source, target) in result.iter().zip(key_guesses.iter_mut()) {
                *target = source.clone();
            }
        }
        for (index, vec) in key_guesses.iter().enumerate() {
            if let Some(&first_byte) = vec.first() {
                found_key[index] = first_byte;
            }
        }
        found_key
    }

    pub fn findkey(&key: &[u8; 16]) -> [u8; 16] {
        let roundkey: [u8; 16] = Self::findroundkey(&key);
        Self::reverse_key_schedule(&roundkey, 4)
    }

    pub fn generate_random_keys(n: usize) -> Vec<[u8; 16]> {
        let mut rng = thread_rng();
        let mut keys_vec = Vec::with_capacity(n);
        let mut key = [0u8; 16];

        for _ in 0..n {
            rng.fill(&mut key);
            keys_vec.push(key);
        }

        keys_vec
    }

    pub fn test_square_attack(n: usize) {
        println!("Square attack against 4-turns AES with {} randomly generated keys", n);
        let keys: Vec<[u8; 16]> = Self::generate_random_keys(n);
        let mut found_key: [u8; 16];
        let mut success: usize = 0;
        for &key in &keys {
            print!("\ngenerated key: ");
            Self::print_key(&key);
            found_key = Aes::findkey(&key);
            if found_key == key {
                print!("found key: ");
                Self::print_key(&found_key);
                success += 1;
            } else {
                println!("key search failed");
            }
        }
        println!("\n{}/{} key(s) found", success, n);
    }

    pub fn print_key(&key: &[u8; 16]) {
        for &byte in &key {
                    print!("{:02x}", byte);
                }
                println!();
    }
}

#[cfg(test)]
mod tests {
    use super::*;
    #[test]
    fn findroundkey_test() {
        let key: [u8; 16] = [
            0x2b, 0x7e, 0x15, 0x16, 0x28, 0xae, 0xd2, 0xa6, 0xab, 0xf7, 0x15, 0x88, 0x09, 0xcf,
            0x4f, 0x3c,
        ];
        let nturn = 4; // doesn't matter for this test
        let aescipher = Aes::new(&key, &nturn);
        let expected: [u8; 16] = aescipher.expanded_key[16..20] // 5th key (1st is the pre_whitenning, 2nd, 3rd and 4th the 3 previous round key)
            .iter()
            .flat_map(|subarray| subarray.iter())
            .copied()
            .collect::<Vec<_>>()
            .try_into()
            .unwrap();
        let found_key: [u8; 16] = Aes::findroundkey(&key);
        assert_eq!(found_key, expected);
    }
    #[test]
    fn reverse_key_schedule_test() {
        let key: [u8; 16] = [
            0x2b, 0x7e, 0x15, 0x16, 0x28, 0xae, 0xd2, 0xa6, 0xab, 0xf7, 0x15, 0x88, 0x09, 0xcf,
            0x4f, 0x3c,
        ];
        let nturn = 4; // doesn't matter for this test
        let aescipher = Aes::new(&key, &nturn);
        let mut round_keys: [[u8; 16]; 11] = [[0u8; 16]; 11];
        for i in 0..11 {
            for j in 0..4 {
                round_keys[i][4 * j..4 * (j + 1)]
                    .copy_from_slice(&aescipher.expanded_key[4 * i + j]);
            }
        }
        for i in 0..11 {
            assert_eq!(Aes::reverse_key_schedule(&round_keys[i], i), key);
        }
    }
    #[test]
    fn find_key_test() {
        let key: [u8; 16] = [
            0x2b, 0x7e, 0x15, 0x16, 0x28, 0xae, 0xd2, 0xa6, 0xab, 0xf7, 0x15, 0x88, 0x09, 0xcf,
            0x4f, 0x3c,
        ];
        assert_eq!(key, Aes::findkey(&key));
    }
}
